/**********************************************************************
 *
 *       lbox.c  -  Routines for making little boxes in BGI.
 *                        Version 1.0
 *
 * NOTICE: Copyright (C) 2000 by Robert S. Larkin
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * The program author can be contacted by email at boblark@proaxis.com.
 * Alternatively contact the author at
 *     Robert S. (Bob) Larkin
 *     2982 N. W. Acacia Place
 *     Corvallis, OR  97330  USA
 *
 * For more details on the background or use of the program, pay a visit to
 * http://www.proaxis.com/~boblark/dsp10.htm  and,
 * http://www.ao.com/~gnome/
 *
 * Initial write 11 Feb 2000 RSL
 */

/* BOX SIZING:
 *   Box sizes should be chosen to have equal top and bottom spaces.
 *   This gives 3 pixels at top and bottom if the first and last lines are
 *   used. Rules, where L is the number of test lines:
 *
 *   Modal boxes should be 15+10*L pixels high (ybottom-ytop), but can
 *   be 1 pixel less and still print the bottom line.
 *
 *   Non-modal boxes have smaller borders and should be 12+10L pixels high.
 *
 *   Each character needs a width of 8 pixels. The rules, where C is number
 *   of characters to fit on a line:
 *
 *   Modal boxes should be 10+8C wide.
 *   Non-modal boxes should be 4+8C wide.
 *
 *   Edit Boxes hold 1 less character.
 *   Radio buttons occupy 16 pixels (2 characters).
 */

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <bios.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#include <graphics.h>
#include <string.h>
#include "lbox.h"

#include "moonsun.h"

#include "uhfa.h"

extern char *big_area_ptr1;
extern char *line_ptr1;

void lb_create(struct lbox_info*, char*, int, int, int, int, int, int);
void lb_create_el(struct edit_info*, int, int, char*, unsigned int, int, int);
void lb_create_rb(struct radio_info*, int, int, int);
void lb_create_check(struct check_info*, int, unsigned int, int);
void lb_edit_state(struct edit_info*, unsigned int);
void lb_radio_active(struct radio_info*, int, unsigned int);
int  lb_open(struct lbox_info*, int);
void lb_update(struct lbox_info*);
int  lb_line(struct lbox_info*, int, int, char*, void*);
unsigned int lb_edit_line(struct lbox_info*, struct edit_info*);
unsigned int lb_radio_button(struct lbox_info*, int, struct radio_info*);
unsigned int lb_check_mark(struct lbox_info*, struct check_info*);
void lb_close(struct lbox_info*, int);

/************************************************************************/

void lb_create(struct lbox_info* lbi, char *nm,
			 int xl, int yt, int xr, int yb, int fst, int mdl)
	{
	int i;

	lbi->xleft=xl;
	lbi->xright=xr;
        lbi->ytop=yt;
        lbi->ybottom=yb;
	strcpy(lbi->name, nm);
	lbi->first_line=fst;
	lbi->modal=mdl;
	if (lbi->modal)
	    {
	    lbi->lines = (yb-yt-14)/10;		/* Num text lines */
	    }
	else
	    {
	    lbi->lines = (yb-yt-11)/10;		/* Num text lines */
	    }
	if (lbi->lines >25)  lbi->lines = 25;
	lbi->open=FALSE;
	lbi->state=0;
	for (i=0;   i < lbi->lines;  i++);
	    {
	    lbi->line_type[i]=0;		/* Empty lines now */
	    lbi->txt[i]=0;			/* ptrs to 0 also */
	    lbi->line_info[i]=0;		/* Ptr to more info structs */
	    }
	}

void lb_create_el(struct edit_info *elptr,
	  int nch, int pos, char *edtxt, unsigned int st, int tpos, int curpos)
	{
	elptr->edit_line_nch=nch;
	elptr->edit_line_pos=pos;
	elptr->edit_line=edtxt;
	elptr->state=st;
	elptr->text_pos=tpos;
	elptr->cursor_pos=curpos;
	}

void lb_create_rb(struct radio_info *rbptr,
	     int bpos, int ipush, int tpos)
	{
	int i;

	rbptr->button_pos=bpos;
	for (i=0; i<25; i++)
	    *( (&rbptr->state[0]) + i)=0X8000;	/* Empty for now */
	*( (&rbptr->state[0]) + ipush)=RADIO_PUSH;
	rbptr->text_pos=tpos;
	}

void lb_create_check(struct check_info *ckptr, int cpos, 
		unsigned int st, int tpos)
	{
	ckptr->check_pos=cpos;
	ckptr->state=st;
	ckptr->text_pos=tpos;
	}

void lb_edit_state(struct edit_info *elptr, unsigned int st)
	{
	elptr->state=st;
	}

void lb_radio_active(struct radio_info *rbptr, int ln, unsigned int st)
	{
	if (st & RADIO_ACTIVE)
	    rbptr->state[ln] |= RADIO_ACTIVE;  /* Act/inact, but not push */
	else rbptr->state[ln] &= 0XFFFD;
	}

int lb_open(struct lbox_info* lbi, int area)
	{
	int xl, xr, yt, yb;
	int bx_ctr, twh;		/* For setting up box title */

	lbi->open=TRUE;

	xl=lbi->xleft;
	xr=lbi->xright;
	yt=lbi->ytop;
	yb=lbi->ybottom;

	setviewport(xl, yt, xr, yb, DO_CLIP);
	if (lbi->modal)			/* Save image to restore at close */
	    {
	    if (imagesize(0, 0, xr-xl, yb-yt) <= 50000)
		{
		if (area==1)
		    getimage(0,  0, xr-xl, yb-yt, big_area_ptr1);
		else
    		    getimage(0,  0, xr-xl, yb-yt, big_area_ptr2);
		}
	    else
		{
		return 0;
		}
	    }

	clearviewport();
	setcolor(c_outline_bold);
	setlinestyle(SOLID_LINE, 0, NORM_WIDTH);
	if (lbi->modal)
	    {
	     rectangle(1, 2, xr-xl-1, yb-yt-1);
	     rectangle(2, 3, xr-xl-2, yb-yt-2);
	     rectangle(4, 5, xr-xl-4, yb-yt-4);
	    }
	else
	    {
	     rectangle(1, 2, xr-xl-1, yb-yt-1);
	    }

	/* Eventually, check size and truncate? */
	settextstyle(DEFAULT_FONT, HORIZ_DIR, 1);
	settextjustify(CENTER_TEXT, TOP_TEXT);

	bx_ctr=(xr+xl)/2;			/* For talking about boxes */
	twh=textwidth(lbi->name)/2;

	/* setviewport is screen coordinates */
	setviewport(bx_ctr-twh-10, yt+1, bx_ctr+twh+10, yt+11, DO_CLIP);
	clearviewport();
	moveto(twh+10, 0);			/* In viewport coordinates */
	outtext(lbi->name);

	return 1;
	}

void lb_update(struct lbox_info* lbi)
	{
	int i, xl, xr, yt, yb;
	int ed_nch;
	int ed_xl, ed_xr;
	int lbl;
	struct edit_info *eiptr;
	struct radio_info *rbptr;
	struct check_info *ckptr;

	xl=lbi->xleft;
	xr=lbi->xright;
	yt=lbi->ytop;
	yb=lbi->ybottom;
	lbl=lbi->lines & 0X001F;	/* The actual line count */

	/* Special case: By setting lbi->lines to 64 over the actual
	 * number of lines, an 80 pixel at the top will be left uncleared
	 * for graphics. (Modal only)
	 */
	if (lbi->modal)
	    {
	    if (lbi->lines & 0X0040)
		{
		setviewport(xl+6, yt+91, xr-5, yb-6, DO_CLIP);
		clearviewport();
		setviewport(xl+6, yt+11, xr-5, yb-6, DO_CLIP);
		}
	    else
		{
		setviewport(xl+6, yt+11, xr-5, yb-6, DO_CLIP);
		clearviewport();
		}
	    }
	else
	    {
	    setviewport(xl+3, yt+11, xr-2, yb-3, DO_CLIP);
	    clearviewport();
	    }

	settextstyle(DEFAULT_FONT, HORIZ_DIR, 1);
	settextjustify(LEFT_TEXT, TOP_TEXT);

	for (i=0;  i < lbl;   i++)
	    {
	    switch (lbi->line_type[i])
		{
		case BLANK:
		   {break;
		   }
		case TXT:
		   {setcolor(c_text);
		    moveto(0, 10*i);
		    outtext(lbi->txt[i]);
		    setcolor(c_text_bold);
		    break;
		   }
		case BOLD:
		   {moveto(0, 10*i);
		    outtext(lbi->txt[i]);
		    break;
		   }
		case EDIT:
		   {/* For edit line_info[i] points to a struct with details
		       of the edit setup */
		    /* Display requires 2 character lines, the designated 
			line and the next down */
		    eiptr=lbi->line_info[i];
		    ed_nch=eiptr->edit_line_nch;
		    ed_xl=eiptr->edit_line_pos;
		    ed_xr=ed_xl+7+8*ed_nch;
		    if (eiptr->state & ED_ACTIVE)
			setcolor(c_text_bold);
		    else
			setcolor(c_text);
		    rectangle(ed_xl, 10*i+3, ed_xr, 10*i+16);
		    moveto(ed_xl+4, 6+10*i);
		    outtext(eiptr->edit_line);
		    moveto(eiptr->text_pos, 6+10*i);
		    outtext(lbi->txt[i]);
		    /* No cursor on in active lines: */
		    setcolor(c_background);
		    line(ed_xl+4, 10*i+14, ed_xr-4, 10*i+14);
		    /* Cursor on active line, beige if at left end */
		    if (eiptr->state & ED_ACTIVE)
			{
			if (eiptr->cursor_pos < eiptr->edit_line_nch-1)
			    setcolor(c_text_bold);
			else
			    setcolor(c_text);
			line(ed_xl+eiptr->cursor_pos*8+4,   10*i+14,
			     ed_xl+eiptr->cursor_pos*8+12,  10*i+14);
			}
		    setcolor(c_text_bold);
		    break;
		   }
		case RADIO:
		   {
		    rbptr=lbi->line_info[i];
		    if (rbptr->state[i] & RADIO_ACTIVE)	/* The active one */
			{
			setcolor(c_text_bold);
			setfillstyle(SOLID_FILL, c_text_bold);
			}
		    else
			{
			setcolor(c_text);
			setfillstyle(SOLID_FILL, c_text);
			}
		    ellipse(9+rbptr->button_pos, 3+10*i, 0, 360, 4, 4);
		    if (rbptr->state[i] & RADIO_PUSH)	/* Pushed */
			{fillellipse(9+rbptr->button_pos, 3+10*i, 2, 2);
			}
		    moveto(rbptr->text_pos, 10*i);
		    outtext(lbi->txt[i]);
		    setcolor(c_text_bold);
		    setfillstyle(SOLID_FILL, c_text_bold);
		    break;
		   }
		case CHECK:
		    {
		    ckptr=lbi->line_info[i];	/* Ptr to check_info struct */
		    if (ckptr->state & CHECK_ACTIVE)	/* If active */
			{
			setcolor(c_text_bold);
			setfillstyle(SOLID_FILL, c_text_bold);
			}
		    else
			{
			setcolor(c_text);
			setfillstyle(SOLID_FILL, c_text);
			}
		    rectangle(5+ckptr->check_pos,   10*i,
			     13+ckptr->check_pos, 8+10*i);
		    if (ckptr->state & CHECK_MARK)	/* Marked */
			{
			line( 6+ckptr->check_pos, 1+10*i,
			     12+ckptr->check_pos, 7+10*i);
			line( 6+ckptr->check_pos, 7+10*i,
			     12+ckptr->check_pos, 1+10*i);
			}
		    moveto(ckptr->text_pos, 10*i);
		    outtext(lbi->txt[i]);
		    setcolor(c_text_bold);
		    setfillstyle(SOLID_FILL, c_text_bold);
		    break;
		    }
		default: break;
		}
	    }
	}

/* lb_line(...)  -  Sets up line, but does not display it. That is
   done by lb_open() and lb_update().
 */
int lb_line(struct lbox_info* lbi, int lnum, int lntype,
		 char* pstr, void* liptr)
	{
	if (lnum > (0X001F & lbi->lines) || lnum<0)
	    return 0;
	lbi->line_type[lnum]=lntype;
	lbi->line_info[lnum]=liptr;
	lbi->txt[lnum]=pstr;
	if (lntype==RADIO)  lb_radio_active(liptr, lnum, RADIO_INACTIVE);

	return 1;
	}

/* lb_edit_line(...)
 *
 * Returns codes, depending on exit character:
 *    Up Cursor:   0X4800
 *    Down Cursor: 0X5000
 *    ENTER:       0X000D
 */
unsigned int lb_edit_line(struct lbox_info* lbi, struct edit_info* el)
	{
	unsigned int c, cl, ch, r;
	int          cli;

	lb_edit_state(el, ED_ACTIVE);
	lb_update(lbi);			/* Make display current */
	while (1)
	   {
	    if ( kbhit() )
		{
		c = _bios_keybrd(_KEYBRD_READ);
		ch=c&0XFF00;
		cl=c&0X00FF;   cli=(int)cl;

		if (cl==ENTER)			/* Enter */
		   {r=cl;
		    break;
		   }
		else if (ch==CUR_UP)		/* Up Cursor */
		   {r=ch;
		    break;
		   }
		else if (ch==CUR_DOWN)		/* Down Cursor */
		   {r=ch;
		    break;
		   }
		else if (cl==BACKSPACE && el->cursor_pos>0)  /* Backspace */
		    {
		    el->cursor_pos--;
		    *(el->edit_line + el->cursor_pos) = 0; /* New EOL */
		    lb_update(lbi);
		    }
		else if ( (el->cursor_pos < el->edit_line_nch-1)
			    && (isgraph(cli) || cli==' ') )
		    {
		    *(el->edit_line + el->cursor_pos) = (char)cl;
		    el->cursor_pos++;
		    *(el->edit_line + el->cursor_pos) = 0; /* New EOL */
		    lb_update(lbi);
		    }

		}     /* End if kbhit() */
	    }         /* End while(1) */
	lb_edit_state(el, ED_INACTIVE);
	return r;
	}

/*************************************************************************
 *
 * lb_radio_button(...)
 *
 * Changes the button state on SPACE
 *
 * Returns codes, depending on exit character:
 *    CUR_UP:    0X4800
 *    CUR_DOWN:  0X5000
 *    ENTER:     0X000D
 **************************************************************************/
unsigned int lb_radio_button
		(struct lbox_info* lbi, int ln, struct radio_info* rl)
	{
	int i;
	unsigned int c, cl, ch, r;

	lb_radio_active(rl, ln, RADIO_ACTIVE);
	lb_update(lbi);			/* Make display current */
	while (1)
	   {
	   if ( kbhit() )
		{
		c = _bios_keybrd(_KEYBRD_READ);
		ch=c&0XFF00;
		cl=c&0X00FF;

		if (ch==CUR_UP) 		/* Up Cursor */
		   {r=ch;
		    break;
		   }
		else if (ch==CUR_DOWN)		/* Down Cursor */
		   {r=ch;
		    break;
		   }
		else if (cl==ENTER)		/* Enter Character */
		   {r=cl;
		    break;
		   }
		else if (cl==SPACE)		/* Change state */
		    {
		    rl->state[ln] |= RADIO_PUSH;	/* Make selection */
		    for (i=0; i<25; i++)		/* Unselect others */
			{
			if (i!=ln && (rl->state[i] != 0X8000) )
			    rl->state[i] &= 0XFFFE;	/* Mark as un-pushed */
			}
		    lb_update(lbi);
		    }
		}     /* End if kbhit() */
	    }         /* End while(1) */
	lb_radio_active(rl, ln, RADIO_INACTIVE);
	return r;
	}

/*************************************************************************
 *
 * lb_check_mark(...)
 *
 * Processing of a check mark line. Changes the check state on SPACE.
 *
 * Returns codes, depending on exit character:
 *    CUR_UP:    0X4800
 *    CUR_DOWN:  0X5000
 *    ENTER:     0X000D
 **************************************************************************/
unsigned int lb_check_mark
		(struct lbox_info* lbi, struct check_info* ck)
	{
	int i;
	unsigned int c, cl, ch, r;

	ck->state |= CHECK_ACTIVE;	/* Highlight this line */
	lb_update(lbi);			/* Make display current */
	while (1)
	   {
	   if ( kbhit() )
		{
		c = _bios_keybrd(_KEYBRD_READ);
		ch=c&0XFF00;
		cl=c&0X00FF;

		if (ch==CUR_UP) 		/* Up Cursor */
		   {r=ch;
		    break;
		   }
		else if (ch==CUR_DOWN)		/* Down Cursor */
		   {r=ch;
		    break;
		   }
		else if (cl==ENTER)		/* Enter Character */
		   {r=cl;
		    break;
		   }
		else if (cl==SPACE)		/* Change state */
		    {
		    ck->state ^= CHECK_MARK;	/* Change mark */
		    lb_update(lbi);
		    }
		}     /* End if kbhit() */
	    }         /* End while(1) */
	ck->state &= 0XFFFD;		/* Make line dim */
	return r;
	}

void lb_close(struct lbox_info* lbi, int area)
	{
	int xl, xr, yt, yb;

	lbi->open=FALSE;

	xl=lbi->xleft;
	xr=lbi->xright;
	yt=lbi->ytop;
	yb=lbi->ybottom;

	setviewport(xl, yt, xr, yb, DO_CLIP);
	if (lbi->modal)			/* Restore image */
	    {
	    if (area==1)
		putimage(0,  0,  big_area_ptr1, COPY_PUT);
	    else
		putimage(0,  0,  big_area_ptr2, COPY_PUT);
	    }
	else
	    {
	    clearviewport();
	    }
	}

/* lb_msg(...)  -  A modal box for display of messages, such as errors.
 * It has a title in the top bar and 2 lines of message room, with the
 * line "Hit ENTER to continue" at the bottom.
 * Exits, and closes box only on Enter.
 * Box size is fixed, but coordinates are from call.
 * Text lines are 30 characters, maximum.
 */
void lb_msg(int xleft, int ytop, char* title, char* ln1, char* ln2)
	{
	struct lbox_info lbmsg;		/* The message box */
	unsigned int c;

	lb_create(&lbmsg, title, xleft, ytop,
		  xleft+250, ytop+63, 0, MODAL);
	lb_open(&lbmsg, 2);
	/* Fixed text in modal box:     */
	lb_line(&lbmsg, 1, BOLD, ln1, 0);
	lb_line(&lbmsg, 2, BOLD, ln2, 0);
	lb_line(&lbmsg, 3, TXT, "    Hit ENTER to continue", 0);
	lb_update(&lbmsg);		/* Make display current */

	/* Now wait for an "Enter" before continuing */
	while (1)
	    {
	    if ( kbhit() )
		{
		c = _bios_keybrd(_KEYBRD_READ);
		if ((c&0X00FF)==ENTER)		/* Enter Character */
		    break;
		}
	    }
	lb_close(&lbmsg, 2);
	}
